/*    Copyright (C) 2003 Dave Ackley                         (-*- C++ -*-)

      This program is free software; you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published by
      the Free Software Foundation; either version 2 of the License, or
      (at your option) any later version.

      This program is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      GNU General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with this program; if not, write to the Free Software
      Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

using namespace std;

#include "Object.h"
#include "World.h"
#include "Parms.h"

Object::_vOp Object::_objects;
size_t Object::_time = 0;
size_t Object::_fishcount = 0;
size_t Object::_sharkcount = 0;
size_t Object::_wandercount = 0;
size_t Object::_centercount = 0;

Object::Object()
  : _x(~0), _y(~0), _mass(0), _idx(-1)
{
  _idx = _objects.size();
  _objects.push_back(this);
}

Object::~Object()
{
  if (_idx >= 0) {
    _objects[_idx] = 0;
  }
}

void Object::UpdateAll(World & w)
{
  _wandercount = 0;
  _centercount = 0;

  // Compact out deaders
  static _vOp temp;

  temp.resize(0);
  for (size_t i = 0; i < _objects.size(); ++i) {
    if (_objects[i])
      temp.push_back(_objects[i]);
  }
  _objects = temp;

  // Shuffle the objects
  size_t max = _objects.size();

  for (size_t i = 0; i < max; ++i) {
    size_t other = random()%max;
    Object * temp = _objects[i];
    _objects[i] = _objects[other];
    _objects[other] = temp;
    
    _objects[i]->_idx = i; 
    _objects[other]->_idx = other; 
  }

  // Iterate over everybody
  for (size_t i = max; i > 0; ) {
    if (Object * o = _objects[--i])
      o->Update(w);
  }

  // Bump the clock
  ++_time;
}

Shark * Object::GetRandomShark() 
{
  Shark * ret = 0;
  size_t scount = 0;
  size_t max = _objects.size();
  for (size_t i = max; i > 0; ) {
    if (Object * o = _objects[--i]) {
      Shark * s = dynamic_cast<Shark*>(o);
      if (s && (random()%++scount)==0) ret = s;
    }
  }
  return ret;
}

bool Fish::Update(World & w)
{
  Parms & p = Parms::Global();

  size_t ex = 0, ey = 0;        // Initialize to avoid compiler warning
  size_t count = 0;

  size_t oldx = _x, oldy = _y;

  for (size_t dir = World::NORTH; dir <= World::WEST; ++dir) {

    size_t cx = oldx, cy = oldy;
    w.GetXY(cx,cy,dir);

    if (w.Get(cx,cy) == 0) {

      if (random() % ++count == 0) {
        ex = cx;
        ey = cy;
      }
    }
  }

  if (count > 0) {

    size_t amt = p.GetFishBite();
    w.AddMass(amt);
    AddMass(amt);

    w.Put(oldx, oldy, 0);
    w.Put(ex, ey, this);
    
    if (GetMass() >= p.GetFishFull()) {
      size_t newmass = (GetMass()+1)/2;
      Fish * f = new Fish();
      RemoveMass(newmass);
      f->AddMass(newmass);

      w.Put(oldx, oldy, f);

    }

  } 

  return true;
}

Shark::Genome::Genome()
{
  // I'm too chicken to believe that all 32 bits out of random() are good.. 
  // and since we only do this code at genesis time, let's go by bits..
  code = 0;
  for (size_t i = 0; i<32; ++i)
    code |= (random()&1)<<i;             
}

Shark::Genome::Genome(const Genome & s)
{
  size_t mutationodds = Parms::Global().GetMutationOdds();
  if (mutationodds < 2) mutationodds = 2;
  code = s.code;
  for (size_t i = 0; i<32; ++i) {
    if (random()%mutationodds == 0)
      code ^= 1<<i;    
  }
}

bool Shark::Update(World & w)
{
  size_t ex = 0, ey = 0;       // Initialize to avoid compiler warning
  size_t ecount = 0;
  size_t fx = 0, fy = 0;       // Initialize to avoid compiler warning
  size_t fcount = 0;

  size_t oldx = _x, oldy = _y;

  Parms & p = Parms::Global();
  size_t burnperscan = p.GetSharkBurn();

  for (size_t idx = 0; idx<Genome::MOVECOUNT; ++idx) {

    size_t dir = _genes.GetMoveCode(idx);

    size_t burned = RemoveMass(burnperscan);
    w.RemoveMass(burned);

    if (GetMass() == 0) {
      w.Put(oldx, oldy, 0);

      delete this;
      return false;
    }

    if (dir == World::CENTER){  // Center means stop scanning now
	++_centercount;
      break;
	}

    if (dir > World::WEST){     // Bogus codes are skipped, but don't stop the scan
      ++_wandercount;
	continue;
	}
    size_t cx = oldx, cy = oldy;
    w.GetXY(cx,cy,dir);

    Object * o = w.Get(cx,cy);
    if (o == 0) {

      if (random() % ++ecount == 0) {
        ex = cx;
        ey = cy;
      }
    } else if (o->IsFish()) {

      ++fcount;
      fx = cx;
      fy = cy;
      break;                    // We quit when we find a fish
    }
  }

  bool moved = false;

  if (fcount > 0) {             // Eating fish at fx,fy

    Object *f = w.Get(fx,fy);

    size_t etmass = f->GetMass();
    f->RemoveMass(etmass);
    AddMass(etmass);

    w.Put(oldx,oldy, 0);
    w.Put(fx,fy,this);
    delete f;
    moved = true;

  } else if (ecount > 0) {      // Moving to empty at ex,ey

    w.Put(ex,ey, this);
    w.Put(oldx, oldy, 0);
    moved = true;
  } 

  if (moved && GetMass() >= _genes.GetBreed()) {
    size_t givamt = _genes.GetToKid();
    if (givamt > GetMass())
      givamt = GetMass();

    Shark * s = new Shark(*this); 
    RemoveMass(givamt);
    s->AddMass(givamt);

    w.Put(oldx, oldy, s);
  }

  return true;

}
